#include <cassert>
#include <cmath>

#include "configg.h"
#include "AuxFuncs.h"
//#include "basrnd.h"
#include "ibtfbase.h"
#include "lightenv.h"
#include "brdfmodel.h"
#include "vrmlsup.h"
#include "tangles.h"
#include "render.h"

// global variables
// the debug counters on the success of BRDF sampling
long long int failBRDFcnt = 0L;
long long int evalBRDFcnt = 0L;
long long int failEMcnt = 0L;
long long int evalEMcnt = 0L;
// counters for a single pixel
int evalEMcntST = 0;
int evalBRDFcntST = 0;
float evalValueVIS = 0;
//void resetCountersST() { evalEMcntST = evalBRDFcntST = 0;}

// ================================================================================
// Evaluating a single point light of white colour
void         // interpolated vertices, tangents, normals and binormals
computeMode0(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF)
{
  float t_i, p_i;
  float result[3];
  // the final value for the pixel
  result[0] = result[1] = result[2] = 0.f;
  // compute the direction of the light for environment map here
  TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
  assert(t_i >= 0.f); assert(t_i <= M_PI);
  
  if (t_i < M_PI/2.0f) {
    float brdfRGB[3];
    // t_i is in radians
    // here we get the common BRDF*cos(theta) value irrespective of the
    // underlying BRDF model
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    const float cosTheta = cos(t_i) + 0.0001f;
    for(int isp=0;isp<=2;isp++) {
      if (smoothTexturing) // smooth texturing illuminated by a lambertian model
	result[isp] = textureImage[isp][irow%textureSize][jcol%textureSize] * cosTheta;
      else
	result[isp] = brdfRGB[isp]; // use BRDF/BTF model only, intensity of light is one
    } // for all channels
  } // if

  // copy the data to the output texture
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = normF * result[isp];
  }  
  setCountersST(0, 1);
  return;
}

// ================================================================================
// Deterministic coherent importance sampling of environment map of
// a single pixel, standard mode either point light or single
// coherent samples at environment map
void         // interpolated vertices, tangents, normals and binormals
computeMode1(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF)
{
  float t_i, p_i;
  float RGB[3], result[3];
  // cout << "A" << std::flush;
  result[0] = result[1] = result[2] = 0.f;
  resetCountersST();
  //cout << " normF = " << normF << endl;
  
  // counting samples
  int failEM = 0;
  int cntAllSamplesEM = 0;
  int cntValidSamplesEM = 0;
  // -----------------------------------------------------------------------------
  // for all samples
  const bool allowNewSamples = false;
  int samplesMax = cntSamplesPerPixel;
  for(int l=0; l < samplesMax; l++) {
    // copy the positions of the directional light for environment map here
    // this direction of light was precomputed
    assert(cntAllSamplesEM < maxCntSamplesPerPixel);
    cntAllSamplesEM++;
    // possibly randomize the access to light
    environMap[actualEnvMap].generateNewRandomLightPrecomputed();
    TA->setIlluDir(environMap[actualEnvMap].getDirectionPrecomputedLight(l)); 
    TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
    // theta and phi are in radians
    if (t_i >= M_PI/2.f) {
      // nothing to do with the illumination from this light
      failEM++;
      if ((allowNewSamples) && (samplesMax < maxCntSamplesPerPixel)) {
	// take additional samples
	samplesMax++;
      }
      continue;
    }
    cntValidSamplesEM++;
    // the value given by BRDF model
    float brdfRGB[3];
    // the cos(theta) factor is already included in the call below
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    // this value has been precomputed
    float emRGB[3];
    for (int isp=0; isp < 3; isp++)
      emRGB[isp] = environMap[actualEnvMap].getColorPrecomputedLight(l, isp);
    float probEM = luminance(emRGB);

    //float probEM = environMap[actualEnvMap].getRGBandProb_ThetaPhi(t_i, p_i, envRGB);
    //float probEM2 = environMap[actualEnvMap].getProb_ThetaPhi(t_i, p_i);
    //float probEM = luminance(envRGB)/(4.0f*M_PI);

    for(int isp=0;isp<=2;isp++) {// for all 3 color channels
      result[isp] +=
	// the colour of light from EM was precomputed
	brdfRGB[isp] * environMap[actualEnvMap].getColorPrecomputedLight(l, isp)/probEM;
    } // for all channels
  } // for all lights
  failEMcnt += failEM;
  evalEMcnt += cntAllSamplesEM;
  // The number of samples taken is cntAllSamplesEM
  // The number of failedSampled is failEM
  // The number of success Samples is cntAllSamplesEM-failEM

  //  float lumEM = luminance(result)/(float)cntAllSamplesEM;
  //  cout << "EM " << lumEM << endl;
  
  float multIntensity = normF / (float)cntAllSamplesEM;
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = result[isp] * multIntensity;
  }
  setCountersST(0, cntAllSamplesEM);
  return;
}

// ================================================================================
// Random importance sampling at environment map, each pixel has a different (unique)
// set of directional lights
void         // interpolated vertices, tangents, normals and binormals
computeMode2(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF)
{
  float result[3];
  result[0] = result[1] = result[2] = 0.f;
  resetCountersST();
  // counting samples
  int failEM = 0;
  int cntAllSamplesEM = 0;
  int cntValidSamplesEM = 0;
  // for all samples
  const bool allowNewSamples = false;
  int samplesMax = cntSamplesPerPixel;
  for(int l=0; l < samplesMax; l++) {
    // set the values to t_l and p_l
    // Now we compute the directional light from the environment map
    float rnd1 = RandomValue();
    float rnd2 = RandomValue();
    float theta, phi; // direction
    // color from the directional light of an environment map
    float envLightColoursRGB[3]; 
    // use the importance sampling to generate light direction and color
    // and probability of selecting this direction, from two random values
    float probEM;
    assert(cntAllSamplesEM < maxCntSamplesPerPixel);
    cntAllSamplesEM++;
    environMap[actualEnvMap].importanceSamplingThetaPhi(rnd1, rnd2, theta, phi, 
							envLightColoursRGB, probEM);
    // the conversion to spherical coordinates
    float XYZ[3];
    ConvertThetaPhiToXYZ(theta, phi, XYZ);
    // set values for illumination, view was set in main function
    TA->setIlluDir(XYZ);
    // convert to the right coordinate system on the object' surface
    float t_i, p_i;
    TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
    // t_i and p_i are in radians
    
    // nothing to do with the light since light aims below the normal
    if (t_i >= M_PI/2.0f) {
      failEM++;
      if ((allowNewSamples) && (samplesMax < maxCntSamplesPerPixel)) {
	// take additional samples
	samplesMax++;
      }
      continue;
    }
    // value given by BRDF, includes already cos(theta) factor
    float brdfRGB[3]; 
    cntValidSamplesEM++;
    // the cos(theta) factor of the rendering equation included in the call below
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    for(int isp=0; isp<=2; isp++) {// for all 3 color channels
      // RGB given by BRDF * RGB given by environment map divided by probability from EM
      result[isp] += brdfRGB[isp] * envLightColoursRGB[isp]/probEM;
    }
  }// for all lights
  // statistics update
  failEMcnt += failEM;
  evalEMcnt += cntAllSamplesEM;
  // The number of all samples taken is 'cntAllSamplesEM'
  // The number of samples that do below surface is 'failEM'
  // The number of samples really taken is 'cntValidSamplesEM'
  
  float multIntensity = normF / (float)cntAllSamplesEM;
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = multIntensity * result[isp];
  }
  setCountersST(0, cntAllSamplesEM);
  return;
}

// ================================================================================
// SAMPLING according to BRDF*cos(theta) or another probability density function
void         // interpolated vertices, tangents, normals and binormals
computeMode3(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol,
	     int BTFindex, float normF)
{
  float result[3];
  result[0] = result[1] = result[2] = 0.f;
  resetCountersST();
  float XYZL[3]; // the direction on the surface of an object
  float envLightColoursRGB[3]; //color from the directional light of an environment map
  // use the importance sampling to get the value
  int cntValidSamplesBRDF = 0;
  int failBRDF = 0;
  int cntAllSamplesBRDF=0;
  const bool allowNewSamples = false;

  for(int l=0; l < cntSamplesPerPixel; l++) {
    // now we compute a ray generated by BRDF model according to BRDF*cos(theta)
    // or some other probability density function that is evaluated to 'probBRDF'
    float rnd1 = RandomValue();
    float rnd2 = RandomValue();
    // compute the stochastic sampling for the two random values
    float probBRDF; // probability of sampling at the direction XYZL
    float brdfRGB[3]; // color=energy of environment map at the direction
    assert(cntAllSamplesBRDF < maxCntSamplesPerPixel);
    cntAllSamplesBRDF++;
    if (brdf->spawnStochasticRay(irow, jcol, rnd1, rnd2, // input
				  t_v, p_v, XYZL, brdfRGB, probBRDF)) { // output
      failBRDF++;
      if (allowNewSamples) {
	if (failBRDF > cntSamplesPerPixel) {
	  BTFimageF[BTFindex] = 0;
	  BTFimageF[BTFindex+1] = 0;
	  BTFimageF[BTFindex+2] = 0;
	  return;
	}
	l--;
      }
      continue; // the ray for rnd1,rnd2 does not exist for the sampling
    }    
    // the number of samples taken to compute the integrand
    cntValidSamplesBRDF++;
    // The result ray in XYZL[3] is in the coordinate system (1,0,0),(0,1,0),(0,0,1)
    // at vertex V, given by vector T,B,N. We convert it to the world-space coordinate
    // system by using T,B,N using standard multiplication of vector by matrix
    float XYZ[3];
    // Normal  0, 0, 1 is converted to Nint
    // Tangent 1, 0, 0 is converted to Tint
    // Bint    0, 1, 0 is converted to Bint
    //
    // This is transpose matrix from local coordinate system to world-space system
    // it is possibly only since vectors T,B,N create orthogonal matrix, then inverse
    // matrix is equal to the transposed matrix
    XYZ[0] = Tint[0]*XYZL[0] + Bint[0]*XYZL[1] + Nint[0]*XYZL[2];
    XYZ[1] = Tint[1]*XYZL[0] + Bint[1]*XYZL[1] + Nint[1]*XYZL[2];
    XYZ[2] = Tint[2]*XYZL[0] + Bint[2]*XYZL[1] + Nint[2]*XYZL[2];
    // Set for the conversion to the global coordinate system
    float t_i, p_i;
    ConvertXYZtoThetaPhi(XYZ, t_i, p_i);
    // the light colour from environment map
    float envRGB[3];
    bool readEM = environMap[actualEnvMap].getRGB_ThetaPhi(t_i, p_i, envRGB);
    for(int isp=0; isp<=2; isp++) // for all 3 color channels
      // we multiply BRDF.cos(theta) from BRDF sampling and RGB colour from EM
      // and divide it by probability from BRDF sampling
      result[isp] += brdfRGB[isp] * envRGB[isp] / probBRDF;
  }// for all samples according to BRDF
  // statistics update of global counters
  failBRDFcnt += failBRDF;
  evalBRDFcnt += cntAllSamplesBRDF;

  // The number of all samples taken is 'cntAllSamplesBRDF'
  // The number of samples that do below surface is 'failBRDF'
  // The number of samples really taken is 'cntValidsamplesBRDF'

  float mult;
  if (cntValidSamplesBRDF)
    mult = normF / (float)cntAllSamplesBRDF; // just dividing by number of samples
  else
    mult = 0.f; // black colour

  //float lumBRDF = luminance(result)/(float)cntAllSamplesBRDF;
  //cout << K << " BRDF " << lumBRDF << endl;
  
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = mult * result[isp];
  }
  setCountersST(cntAllSamplesBRDF, 0);
  return;
}


