// ================================================================
// $Id: $
//
// render4.cpp
//
// This implements MIS, combining EM and BRDF*cos(theta) sampling, without storing the samples
// MIS = Multiple Importance Sampling (Veach and Guibas 1995)
//
// Initial coding by Vlastimil Havran, 2013.

#include <cassert>
#include <cmath>

#include "configg.h"
#include "AuxFuncs.h"
#include "ibtfbase.h"
#include "lightenv.h"
#include "brdfmodel.h"
#include "vrmlsup.h"
#include "tangles.h"
#include "render.h"

// Simple correct implementation of MIS in deterministic way
// without storing samples and probabilities to the array
//
void         // interpolated vertices, tangents, normals and binormals
computeMode4(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol,
	     int BTFindex, float normF)
{
  double result[3];
  result[0] = result[1] = result[2] = 0.f;
  resetCountersST();
  float XYZL[3]; // the direction on the surface of an object
  float envLightColoursRGB[3]; //color from the directional light of an environment map
  // use the importance sampling to get the value

  const bool fixNumberOfSamples = false;// either fix it or not
  int cntSamplesPerPixelBRDF = (1.0f-alphaMIS) * (float)cntSamplesPerPixel;
  int cntSamplesPerPixelEM = cntSamplesPerPixel - cntSamplesPerPixelBRDF;
  // the computed alpha can be different due to the discretization
  float alphaMISreal = (float)cntSamplesPerPixelBRDF/(float)cntSamplesPerPixel;
  int cntValidSamplesBRDF = 0; // the number of samples really taken from BRDF
  int detectGetStucked = 0;
  int cntAllSamplesBRDF = 0;
  // compute the samples from BRDF
  for(int l=0; l < cntSamplesPerPixelBRDF; l++) {
    // now we compute a ray generated by BRDF model according to BRDF*cos(theta)
    // or some other probability density function that is evaluated to 'probBRDF'
    float rnd1 = RandomValue();
    float rnd2 = RandomValue();
    // compute the stochastic sampling for the two random values
    float probBRDF; // probability of sampling at the direction XYZL
    float brdfRGB[3]; // color=energy of environment map at the direction
    cntAllSamplesBRDF++;
    if (brdf->spawnStochasticRay(irow, jcol, rnd1, rnd2, // input
				 t_v, p_v, XYZL, brdfRGB, probBRDF)) { // output
      detectGetStucked++;
      if (fixNumberOfSamples) {
	if (detectGetStucked > cntSamplesPerPixelBRDF*2) {
	  BTFimageF[BTFindex] = 0;
	  BTFimageF[BTFindex+1] = 0;
	  BTFimageF[BTFindex+2] = 0;
	  return;
	}
	l--; // we have to really taken specified number of samples
      }
      // so we repeat the sampling
      continue; // the ray for rnd1,rnd2 does not exist for the sampling
    }
    // The result ray in XYZL[3] is in the coordinate system (1,0,0),(0,1,0),(0,0,1)
    // at vertex V, given by vector T,B,N. We convert it to the world-space coordinate
    // system by using T,B,N using standard multiplication of vector by matrix
    float XYZ[3];
    // Normal  0, 0, 1 is converted to Nint
    // Tangent 1, 0, 0 is converted to Tint
    // Bint    0, 1, 0 is converted to Bint
    //
    // This is transpose matrix from local coordinate system to world-space system
    // it is possibly only since vectors T,B,N create orthogonal matrix, then inverse
    // matrix is equal to the transposed matrix
    XYZ[0] = Tint[0]*XYZL[0] + Bint[0]*XYZL[1] + Nint[0]*XYZL[2];
    XYZ[1] = Tint[1]*XYZL[0] + Bint[1]*XYZL[1] + Nint[1]*XYZL[2];
    XYZ[2] = Tint[2]*XYZL[0] + Bint[2]*XYZL[1] + Nint[2]*XYZL[2];
    // Set for the conversion to the global coordinate system
    float t_i, p_i;
    ConvertXYZtoThetaPhi(XYZ, t_i, p_i);
    // the light colour from environment map
    float envRGB[3]; // and the probability of sampling
    float probEM = environMap[actualEnvMap].getRGBandProb_ThetaPhi(t_i, p_i, envRGB);
    // here we compute weight for MIS heuristics, the first parameters determines if the weight
    // is for BRDF or EM, now it is for BRDF
    float weight = evaluateMISweight(probBRDF, cntSamplesPerPixelBRDF, probEM, cntSamplesPerPixelEM);
    
    for(int isp=0; isp<=2; isp++) // for all 3 color channels
      // we multiply BRDF.cos(theta) from BRDF sampling and RGB colour from EM
      // and divide it by probability from BRDF sampling
      result[isp] += weight * brdfRGB[isp] * envRGB[isp] / probBRDF;

    // the number of samples taken to compute the integrand
    cntValidSamplesBRDF++;
  }// for all samples according to BRDF
  // global counters
  failBRDFcnt += detectGetStucked;
  evalBRDFcnt += cntAllSamplesBRDF;

  // intensity multiplier
  float mult;
  mult = 1.0f / (float)cntAllSamplesBRDF; 
  for(int isp=0;isp<=2;isp++) {
    // here we assign the result
    BTFimageF[BTFindex+isp] = mult * result[isp];
  }
  // we have the result in BTFimageF for BRDF sampling

  // -----------------------------------------------------------------------------
  // Sampling according to EM
  result[0] = result[1] = result[2] = 0; // reset the sum again
  int cntValidEMSamples = 0;
  int cntAllSamplesEM = 0;
  int samplesMax = cntSamplesPerPixelEM;
  int failsEM = 0;
  for(int l=0; l < samplesMax; l++) {
    // copy the positions of the directional light for environment map here
    // this direction of light was precomputed.
    // possibly randomize direction lights
    environMap[actualEnvMap].generateNewRandomLightPrecomputed();
    TA->setIlluDir(environMap[actualEnvMap].getDirectionPrecomputedLight(l)); 
    float t_i, p_i;  // direction (theta, phi) from the environment map
    cntAllSamplesEM++;
    assert(cntAllSamplesEM < maxCntSamplesPerPixel);
    TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
    if (cntAllSamplesEM == maxCntSamplesPerPixel) {
      cout << "Error, the number of precomputed lights were exhausted without"
	   << " finding the required number of samples = " << cntSamplesPerPixelEM << endl;
      BTFimageF[BTFindex+0] = MARK_PROBLEM; // make pixel RED
      BTFimageF[BTFindex+1] = 0;
      BTFimageF[BTFindex+2] = 0;
      cout << "RED" << flush;
      return; // and return result
    }     
    // theta and phi are in radians
    if (t_i >= M_PI/2.f) {
      failsEM++; // the number of failures for EM sampling
      if (fixNumberOfSamples)
	samplesMax++;
      // nothing to do with the illumination from this light direction, continue sampling
      continue;
    }

    // the value given by BRDF model
    float brdfRGB[3];
    // the cos(theta) factor is already included in the call below
    // including the probability of sampling from BRDF*cos(theta)
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    // compute probability of sampling by current algorithm
    float probBRDF = brdf->getProbOfSampling(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);

    // probability of sampling to some directions can be zero, since if we have
    // only specular function or some parts with zero values, this will exactly happen
    // in physical reality impossible, in mathemathical model easily possible
    if (probBRDF <= 0.f) {
      probBRDF = luminance(brdfRGB) * 1e10f+1e-20; // so the result will be zero as value/prob is the result
    }
    assert(probBRDF > 0.f);

    // this value has been precomputed 
    float probEM = environMap[actualEnvMap].getProbPrecomputedLight(l);    
    // here we compute weight for MIS heuristics, the first parameters determines if the weight
    // is for BRDF or EM, now it is weight for EM
    float weight = evaluateMISweight(probEM, cntSamplesPerPixelEM, probBRDF, cntSamplesPerPixelBRDF);
    for(int isp=0;isp<=2;isp++) {// for all 3 color channels
      result[isp] +=
	// the colour of light from EM was precomputed
	weight * brdfRGB[isp] * environMap[actualEnvMap].getColorPrecomputedLight(l, isp)/probEM;
    } // for all channels
  } // for all lights
  // global counters
  failEMcnt += failsEM;
  evalEMcnt += cntAllSamplesEM;

  // we sum the result in BTFimageF for BRDF sampling and 'result' for EM sampling  
  float multIntensity = normF / (float)cntAllSamplesEM;  
  for(int isp=0;isp<=2;isp++) {
    // here we add the result, since it is not empty
    BTFimageF[BTFindex+isp] += result[isp] * multIntensity;
  }
  
  setCountersST(cntAllSamplesBRDF, cntAllSamplesEM);
  return;
}

